// $RCSfile: MPClient.java,v $
// $Author: aboyer $ 
// $Date: 2004/02/25 06:02:53 $
// $Revision: 1.17 $

package com.spamnbeans.client;

import com.spamnbeans.server.*;

import javax.naming.InitialContext;
import javax.naming.Context;
import javax.rmi.PortableRemoteObject;

import java.util.logging.*;
import java.io.*;
import java.util.Hashtable;

/**
 * Spam-n-Beans mail processing system client interface.
 * <p>
 * This client interface connects to a Mail Processing Server and submits an
 * email for processing.  The email is read from standard input, and a copy of
 * the email message is printed on standard output along with any relevant
 * changes.  It will attempt to retry several times if errors occur.
 *
 * @author Andrew Boyer,
 *			<a href="mailto:aboyer@andrew.cmu.edu">aboyer@andrew.cmu.edu</a>
 * @version 1.0
 *
 * @author Philippe M. Wilson
 *			&lt;<a href="mailto:pmwilson@cmu.edu">pmwilson@cmu.edu</a>&gt;
 * @version 1.2
 */
public class MPClient
{
	private static Logger logger = Logger.getLogger(MPClient.class.getName());
	
	private static ManagerHomeRemote mgrhome; 
	private static ManagerRemote mgr;

	private static MailProcHomeRemote mphome;
	private static MailProcRemote mp;

	public static void main(String[] args){// throws Throwable
		BufferedReader in;
		PrintWriter out;
		String[] msg;

		// check args, determine settings
		// <none for now>

		// setup in/out with stdin/out
		in = new BufferedReader(new InputStreamReader(System.in));
		out = new PrintWriter(System.out);

		// get message from 'in'
		msg = getMessage(in);
		String m = msg[0] + msg[1];
		if (m.length() == 0) {
			logger.severe( "Input message cannot be of length 0");
			System.exit(220);
		}
		// connect to naming service
		connectNamingService();

		// connect to a processing server
		connectProcessingServer();


		// send message to the server for processing
		try {
			String result = mp.processMessage( new Integer( args[0] ),
					new Integer(m.hashCode()), m);
			System.out.print(msg[0] + result + msg[1]);
		} catch (InvalidClientException ice) {  // Client rejected
			logger.severe("Invalid Client ID");
			System.exit(230);
		} catch (Throwable t) {
			if (logger.isLoggable(Level.FINEST)) {
				logger.log(Level.FINEST, "Error processing message:", t);
			}
			else {
				logger.severe("Failed to process message");
			}
			System.exit(100);
		}


	}

	/**
	 * Create Naming Service connection.
	 */
	private static void connectNamingService(){		
		logger.info("Connecting to Naming Service...");
		try {
			Context jndiContext = new InitialContext();

			Object mgrref = jndiContext.lookup("ManagerHomeRemote");
		
			logger.info("Connected to Naming Service (JNDI)");
			
			mgrhome = (ManagerHomeRemote)
				PortableRemoteObject.narrow(mgrref,ManagerHomeRemote.class);
		} catch (Throwable t) {
			if (logger.isLoggable(Level.FINEST)) {
				logger.log(Level.FINEST, "Error connecting to Naming Service:", t);
			} else {
				logger.warning("Error connecting to Naming Service");
			}
			logger.severe("Failed to connect to the Naming Service");
			System.exit(100);
		}
	}


	/**
	 * Create Processing Server connection.
	 */
	private static void connectProcessingServer(){		
		logger.info("Connecting to a Processing Server...");
		boolean successMPServer=false;
		for(int i=0; i<8 && successMPServer==false; i++){

			// connect to replication manager
			boolean successRepMan=false;
			while(successRepMan==false){
				try {
					mgr = mgrhome.create();
					
					logger.info("Connected to Load Balancer (Replication Manager)");
					successRepMan=true; // DONE, move on to next step
				} catch (Throwable t) {
					if (logger.isLoggable(Level.FINEST)) {
						logger.log(Level.FINEST, "Error connecting to Load Balancer (Replication Manager)", t);
					} else {
						logger.warning("Error connecting to Load Balancer (Replication Manager)");
					}
					break; // don't retry
				}
			}
			if(successRepMan==false)
				continue; // retry
			
			// connect to replicated processing server
			boolean successRepMPServer=false;
			for(int j=0; j<2 && successRepMPServer==false; j++){ // try a few times
				try {
					Integer mpid = mgr.serverLoadBalance();
					String mpname = mgr.serverGetName( mpid );

					logger.info("Directing request to "+mpname+" ...");

					Context jndiContext = new InitialContext();
					jndiContext.addToEnvironment("java.naming.provider.url", mpname);
					Object mpref = jndiContext.lookup("MailProcHomeRemote");

					logger.info("Connecting to Processing Server...");

					mphome = (MailProcHomeRemote)
						PortableRemoteObject.narrow(mpref,MailProcHomeRemote.class);

					mp = mphome.create();
					mp.attachServer(mpid);

					logger.info("Connected to Processing Server: " + mp.getServerName());
					successRepMPServer=true; // DONE, move on to next step
				} catch (Throwable t) {
					if (logger.isLoggable(Level.FINEST)) {
						logger.log(Level.FINEST, "Error connecting to Processing Server", t);
					} else {
						logger.warning("Error connecting to Processing Server");
					}
				}
			}
			if(successRepMPServer==false)
				continue; // retry

			successMPServer=true; // successful connection to a Mail Processing Server
		}
		if(successMPServer==false){
			logger.severe("Failed to connect to a Processing Server");
			System.exit(110);

		}
	}
	
	/**
	 * Read a message from standard input.
	 */
	private static String[] getMessage(BufferedReader in){		
		logger.info("Getting message...");

		String header = null;
		String body = null;

		try {
			String inputLine;
			StringBuffer buf = new StringBuffer();
			while ((inputLine = in.readLine()) != null) {
				if (header == null && inputLine.length() == 0) {
					header = buf.toString();
					buf = new StringBuffer();
				}
				buf.append(inputLine);
				buf.append("\r\n");
			}
			body = buf.toString();

			in.close();

			if (header == null) {
				// pretend like entire message is header
				header = body;
				body = "";
			}
		} catch (Throwable t) {
			if (logger.isLoggable(Level.FINEST)) {
				logger.log(Level.FINEST, "Error reading input message:", t);
			}else {
				logger.severe("Failed to read input message");
			}
			System.exit(210);
		}

		String[] rtn = new String[2];
		rtn[0]=header;
		rtn[1]=body;
		return rtn;
	}

}
