// $Id: ReplicaMonitor.java,v 1.25 2004/03/21 23:02:17 pmwilson Exp $

package com.spamnbeans.repmgr;

//import com.spamnbeans.client.CrashFaultInjector;
import com.spamnbeans.server.*;

import java.util.Properties;
import java.io.IOException;
import java.rmi.RemoteException;

import javax.naming.InitialContext;
//import javax.naming.Context;
import javax.naming.NamingException;
import javax.rmi.PortableRemoteObject;
import javax.ejb.CreateException;

import java.util.logging.*;

import java.util.*;
import java.io.*;

/*
 * This is needed to consume & redirect output from exec processes
 * - Failure to do this could cause the thread to deadlock if
 *   it produces output when no-one is listening...
 */
class StreamGobbler extends Thread
{
    InputStream is;
    String type;
    OutputStream os;

    StreamGobbler(InputStream is, String type)
    {
	this(is, type, null);
    }

    StreamGobbler(InputStream is, String type, OutputStream redirect)
    {
	this.is = is;
	this.type = type;
	this.os = redirect;
    }

    public void run()
    {
	try {
	    PrintWriter pw = null;
	    if( os != null ) {
		pw = new PrintWriter(os);
	    }
	    InputStreamReader isr = new InputStreamReader(is);
	    BufferedReader br = new BufferedReader(isr);
	    String line = null;
	    while( (line = br.readLine()) != null ) {
		if( pw != null ) {
		    pw.println(line);
		}
		System.out.println(type + ">" + line);
	    }
	    if( pw != null ) {
		pw.flush();
	    }
	} catch( IOException e) {
	    e.printStackTrace();
	}
    }
}

/**
 * The Replica Monitor is a thread that is devoted to monitoring a replica.
 * <p>
 * The thread also obtains server utilization statistics for load balancing.
 * 
 * @author Gary Ackley, <a href="mailto:gca@andrew.cmu.edu">gca@andrew.cmu.edu
 *         </a>
 * @version $Revision: 1.25 $
 */
public class ReplicaMonitor extends Thread { // Runnable
	private static Logger logger =
		Logger.getLogger(GlobalReplicationManager.class.getName());

	// Data passed in by constructor
	private int serverWatchedID;
	private String serverName;
	private int sleepPeriod = 60;
	private GlobalReplicationManager grm = null;
	private ManagerRemote mgr = null;

	private Thread MonitorThread = null;

	private static final int maxPingFailures = 4;
	private int pingFailures = maxPingFailures;

	// The handle to the ReplicaState
	private ReplicaStateRemote replicaState = null;
	private boolean isAlive = false;
	private boolean isPolling = true;
	private long numClients = 0;
	private Process psJBossStart = null;

	/**
	 * Create a new ReplicaMonitor Object. <br>Note: Monitor is not running
	 * until caller explicitly calls "run"
	 * 
	 * @param serverID
	 * @param name
	 * @param checkPeriod
	 * @param grm
	 */
	public ReplicaMonitor(
		int serverID,
		String name,
		int checkPeriod,
		GlobalReplicationManager grm,
		ManagerRemote mgr) {

		// Set name of thread
		super(name);

		String strMsg = "Created ReplicaMonitor[" + serverID + "]:" + name;
		System.out.println(strMsg);
		logger.info(strMsg);

		this.serverWatchedID = serverID;
		this.sleepPeriod = checkPeriod;
		this.serverName = name;
		this.grm = grm;
		this.mgr = mgr;
	}

	/**
	 * Watches the server. <br>If the ping results in an exception the server
	 * is shutdown & restarted
	 */
	public void run() {
		String strMsg =
			"ReplicaMonitor launched for ServerID["
				+ this.serverWatchedID
				+ "]:"
				+ this.serverName;
		logger.info(strMsg);

		// Save this for later...
		MonitorThread = currentThread();

		// Enter an infinite loop, pinging the server until
		// the GlobalReplicationManager tells us to stop (or we
		// get interrupted...)
		while (isPolling) {
			try {

				// Get handle to replica if needed.
				if (replicaState == null) {
					this.replicaState = bindServer();
				}

				// Ping the replica
				// strMsg = "pinging server";
				// logger.log(Level.INFO, strMsg);
				// replicaState.ping();
				numClients = replicaState.getNbrRequests();
				strMsg = "Server["
				    + this.serverWatchedID
				    + "]:" + this.serverName
				    + " has "
				    + numClients
				    + " clients attached.";
				logger.log(Level.INFO, strMsg);

				pingFailures = 0; // reset ping failure count
				
				try {
					mgr.serverSetEnabled(new Integer(serverWatchedID), true); // enable use
				} catch (Exception srv) {
					isPolling = false;
				}
			} catch (Exception e) {
				// TODO: short-circuit by checking if a client has already 
				// disabled the server.  if so, go right ahead w/ recovery 

				pingFailures++;

				strMsg =
					"Server["
					+ this.serverWatchedID
					+ "]: "
					+ this.serverName
					+ " has not responded to " + pingFailures + " pings.";
				logger.log(Level.WARNING, strMsg);

				if(pingFailures >= maxPingFailures){
					// !!!Server is dead!!!!
					// Do something about it
					strMsg =
						"Server["
						+ this.serverWatchedID
						+ "]: "
						+ this.serverName
						+ " is dead!";
					logger.log(Level.WARNING, strMsg);
					try {
						mgr.serverSetEnabled(new Integer(serverWatchedID), false); // disable use
					} catch (Exception srv) {
						isPolling = false;
					}
					haltServer();
					replicaState = null;
					startServer();
				}
			}

			// Give the Remote Replica a chance to do something useful
			try {
				Thread.sleep(1000 * this.getCheckPeriod());
			} catch (InterruptedException ie) {
				// swallow the interruption
				strMsg = "Caught InterruptedException on " + serverName;
				logger.log(Level.INFO, strMsg);
			}
		}

		strMsg =
			"ReplicaMonitor terminated for ServerID["
				+ this.serverWatchedID
				+ "]:"
				+ this.serverName;
		logger.log(Level.INFO, strMsg);
	}

	private void stopManager() {
		isPolling = false;
	}

	public void halt() {
		//halt the server
		this.stopManager();
		haltServer();

		// Wake up MonitorThread, in case it's sleeping
		if (MonitorThread.isAlive()) {
			MonitorThread.interrupt();
		}
	}

	/**
	 * Binds to the remote server
	 * 
	 * @return Success/Failure of binding to remote server
	 * @throws ServerException
	 */
	private ReplicaStateRemote bindServer() throws ServerException {
		ReplicaStateRemote rs = null;
		String strObjName = "ReplicaStateHomeRemote";
		String strMsg = "ReplicaMonitor:bindServer";
		logger.log(Level.FINE, strMsg);

		/*
		 * javax.naming.Context represents a naming context, which is used to
		 * look up and manage bindings and subcontexts. javax.naming.Name
		 * provides an abstract representation of a name in a naming service.
		 * javax.naming.Binding is a representation of a naming service name
		 * and the object to which the name is bound. javax.naming.Reference is
		 * a representation of how to get a copy of an object.
		 */

		// Construct a name consisting of ReplicaState_<sid>
		//String homeName = strObjName + "_" + this.serverWatchedID;
		//strMsg = "Binding to local object:" + homeName;
		//logger.info(strMsg);

		// Lookup the object in the naming service on the remote machine.
		// - If the lookup fails, then the replica hasn't been
		//   registered yet (or has failed) needs to be restarted
		// - If this works, then the replica has started and we can
		//   connect to the ReplicaState for that replica to monitor it.
		try {
			Properties props = System.getProperties();
			InitialContext jndiContext = new InitialContext(props);
			jndiContext.addToEnvironment(
				"java.naming.provider.url",
				this.serverName);
			strMsg = "Connecting to naming service on: " + this.serverName;
			logger.info(strMsg);
			//			Object objRef = jndiContext.lookup(homeName);
			Object objRef = jndiContext.lookup(strObjName);
			ReplicaStateHomeRemote home =
				(ReplicaStateHomeRemote) PortableRemoteObject.narrow(
					objRef,
					ReplicaStateHomeRemote.class);
			rs = home.create();
			this.isAlive = true;
			strMsg =
				"Server["
					+ this.serverWatchedID
					+ "]:"
					+ this.serverName
					+ " is alive!";
			logger.info(strMsg);
		} catch (javax.naming.CommunicationException e) {
			// if naming exception encountered with the Context, such
			// as when JNDI server is down.
			strMsg =
				"Caught CommunicationException - JNDI server unresponsive!";
			// logger.log(Level.WARNING, strMsg, e);
			// logger.log(Level.WARNING, strMsg);
			throw new ServerException(strMsg);
		} catch (NamingException e) {
			// if naming exception encountered with the Context, such
			// as when naming service could not find object.
			strMsg = "Caught NamingException! " + strObjName;
			// logger.log(Level.WARNING, strMsg, e);
			// logger.log(Level.WARNING, strMsg);
			throw new ServerException(strMsg);
		} catch (ClassCastException e) {
			// ClassCastException if PortableRemoteObject cannot cast to
			// UPSessionHome interface
			strMsg = "Caught ClassCastException! " + strObjName;
			//			logger.log(Level.WARNING, strMsg, e);
			// logger.log(Level.WARNING, strMsg);
			throw new ServerException(strMsg, e);
		} catch (CreateException e) {
			// CreateException if exception encountered while creating the bean
			strMsg = "Caught CreateException! " + strObjName;
			//			logger.log(Level.WARNING, strMsg, e);
			// logger.log(Level.WARNING, strMsg);
			throw new ServerException(strMsg, e);
		} catch (RemoteException e) {
			// RemoteException if exception encountered while connecting to
			// server
			strMsg = "Caught RemoteException! " + strObjName;
			//			logger.log(Level.WARNING, strMsg, e);
			// logger.log(Level.WARNING, strMsg);
			throw new ServerException(strMsg, e);
		}
		
		// Return the object to the caller
		return rs;		
	}

	/**
	 * Start the server (JBoss) remotely via SSH <br>This assumes
	 * <ul>
	 * <li>$JBOSS_HOME is set for user on remote systems
	 * <li>remote server already has the correct copy of the JAR files
	 * deployed (or has remotely mounted the file system that contains them).
	 * </ul>
	 */
	private void startServer() {
		String strMsg = "ReplicaMonitor:startServer";
		logger.log(Level.FINE, strMsg);

		try {
			strMsg =
				"starting Server["
					+ this.serverWatchedID
					+ "]:"
					+ this.serverName;
			//String cmd = "$JBOSS_HOME/bin/run.sh -c team3 2>&1 >/tmp/JBoss.log";
			String cmd = "'$JBOSS_HOME/bin/run.sh -c team3'";

			// For windows systems, must prefix with "cmd /c"
			// - Note: Still using same command, assuming cygwin is installed
			String[] args = new String[] { cmd };
			String os = System.getProperty("os.name");
			if (os.startsWith("Windows")) {
				args = new String[] {
				    "cmd", "/c",
				    "C:\\cygwin\\bin\\ssh",
				    serverName,
				    " \"" + cmd	+ "\""
				};
			} else if (os.startsWith("Linux")) {
			    args = new String[] {
				// "/bin/sh", "-c", 
				// "$SPAM_REPOS/scripts/repman",
				"./scripts/repman",
				"-p" , cmd,
				"-l" , "Replica_" + serverName ,
				"-m" , serverName
			    };
			}

			if (this.serverName.equalsIgnoreCase("localhost")) {
				// Do not mess with console system or bad things
				// could happen...
				strMsg = "Avoiding starting server on localhost";
				logger.info(strMsg);
			} else {
				strMsg = strMsg + "\n\tOS:" + os + " Exec: ";
				for (int i = 0; i < args.length; i++) {
				    strMsg += " "+args[i];
				}
				logger.info(strMsg);
				psJBossStart = Runtime.getRuntime().exec(args);
				StreamGobbler stderr = new StreamGobbler(psJBossStart.getErrorStream(),
									 "ERROR");
				StreamGobbler stdout = new StreamGobbler(psJBossStart.getInputStream(),
									 "OUTPUT");
				stderr.start();
				stdout.start();
				strMsg = "Waiting for " + serverName + " to wake up...";
				logger.log(Level.INFO, strMsg);
				sleep(1000 * 60);
				strMsg = serverName + " should be awake by now.";
				logger.info(strMsg);
			}

		} catch (InterruptedException e) {
			logger.log(Level.WARNING, "startServer interrupted: ", e);
		} catch (IOException e) {
			strMsg = "Exec failed - Terminating ReplicaMonitor";
			logger.log(Level.SEVERE, strMsg, e);
			this.isPolling = false;
		}

		// If JBoss start successful - then deploy Replica by
		// uploading to remote server...
		deployReplica();
	}

	/**
	 * Uploads the JAR file to the server for deployment
	 */
	private void deployReplica() {
		String strMsg = "ReplicaMonitor:deployReplica";
		logger.log(Level.FINE, strMsg);

		try {
			String deploydir = "$JBOSS_HOME/server/team3/deploy";
			String cmd =
				"scp $deploydir/spamnbeans.jar "
					+ this.serverName
					+ ":$deploydir";

			// For windows systems, must prefix with "cmd /c"
			// - Note: Still using same command, assuming cygwin is installed
			String os = System.getProperty("os.name");
			if (os.startsWith("Windows")) {
				cmd = "cmd /c C:\\cygwin\\bin\\" + cmd;
			} else {
				// Don't do this for Linux Clusters...
				// - they already point to same deployment dir.
				return;
			}

			if (this.serverName.equalsIgnoreCase("localhost")) {
				// Do not mess with console system or bad things
				// could happen...
				strMsg = "Avoiding deployment on localhost";
				logger.info(strMsg);
			} else {
				strMsg =
					"Uploading Replica to Server["
						+ this.serverWatchedID
						+ "]:"
						+ this.serverName;
				logger.info(strMsg);
				logger.log(Level.INFO, "Exec:" + cmd);
				Process process = Runtime.getRuntime().exec(cmd);
				strMsg = "Waiting for deployment to finish...";
				logger.log(Level.INFO, strMsg);
				sleep(1000 * 5); // give it 5 seconds
				process.waitFor();
				strMsg = "Replica deployed on:" + serverName;
				logger.info(strMsg);
			}

		} catch (InterruptedException e) {
			logger.log(Level.WARNING, "startServer interrupted: ", e);
		} catch (IOException e) {
			strMsg = "Exec failed - Terminating ReplicaMonitor";
			logger.log(Level.SEVERE, strMsg, e);
			this.isPolling = false;
		}
	}

	/**
	 * Terminates a server (JBoss) remotely
	 */
	private void haltServer() {
	    // There are several ways to terminate a replica:
	    //  1) use the psJBoss handle to kill the replica process
	    //  2) use "repman -k <label>"
	    //  3) invoke "shutdown.sh" via SSH
	    //  4) invoke "shutdown.sh" via repman
	    // - 1 & 2 are very unfriendly ways of shutting down the remote 
	    //   process & is best avoided if possible.
	    // - 3 & 4 are arguably the same thing, and is the nicest
	    //   and cleanest way to shutdown JBoss.
	    // - note that 3 avoids creating a very short-lived 'screen'.
	    // - 4 is used here since it uses the same mechanism that
	    //   works in startServer
	    String strMsg = "ReplicaMonitor:haltServer";
	    logger.log(Level.FINE, strMsg);

		try {
		    strMsg =
			"Terminating Server["
			+ this.serverWatchedID
			+ "]:"
			+ this.serverName
			+ "...";
			String cmd = "'$JBOSS_HOME/bin/shutdown.sh -S'";
			String[] args = new String[] { cmd };
			String os = System.getProperty("os.name");
			if (os.startsWith("Windows")) {
				args = new String[] {
				    "cmd", "/c",
				    "C:\\cygwin\\bin\\ssh",
				    serverName,
				    cmd
				};
			} else if (os.startsWith("Linux")) {
			    args = new String[] {
				// "/bin/sh", "-c", 
				// "$SPAM_REPOS/scripts/repman",
				"./scripts/repman",
				"-p" , cmd,
				"-l" , "ReplicaKill_" + serverName ,
				"-m" , serverName
			    };
			}

			if (this.serverName.equalsIgnoreCase("localhost")) {
				// Do not mess with console system or bad things
				// could happen...
				strMsg = "Avoiding killing server on localhost";
				logger.info(strMsg);
			} else {
				this.isAlive = false;
				strMsg = strMsg + "\n\tOS:" + os + " Exec: ";
				for (int i = 0; i < args.length; i++) {
				    strMsg += " "+args[i];
				}
				logger.info(strMsg);
				Process process = Runtime.getRuntime().exec(args);
				StreamGobbler stderr = new StreamGobbler(process.getErrorStream(),
									 "ERROR");
				StreamGobbler stdout = new StreamGobbler(process.getInputStream(),
									 "OUTPUT");
				// Wait for original SSH thread to terminate
				if (psJBossStart == null)
					return;
				strMsg = "Waiting for " + serverName + " to die...";
				logger.log(Level.INFO, strMsg);
				sleep(1000 * 5); // give it 5 seconds
				process.waitFor();
				psJBossStart.waitFor(); // Wait for process to exit
				strMsg = serverName + " has terminated.";
				logger.info(strMsg);
			}

		} catch (InterruptedException e) {
			logger.log(Level.WARNING, "startServer interrupted: ", e);
		} catch (IOException e) {
			strMsg = "Exec failed - Terminating ReplicaMonitor";
			logger.log(Level.SEVERE, strMsg, e);
			this.isPolling = false;
		}
	}

	public int getCheckPeriod() {
		return this.sleepPeriod;
	}

	public void setCheckPeriod(int newperiod) {
		this.sleepPeriod = newperiod;
	}

	public long getNumClients() {
		return this.numClients;
	}

	public String getServerName() {
		return this.serverName;
	}

	public int getServerId() {
		return this.serverWatchedID;
	}

	public ReplicaStateRemote getReplicaState() {
		return replicaState;
	}
}
